unit DirectSpeechSynthesisControlU;

//Uses the low level MS Direct Speech Synthesis ActiveX (ActiveVoice)
//to simplify coding and to provide animation

interface

uses
  Speech,
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, ImgList, StdCtrls, ComCtrls, ToolWin, OleCtrls,
  ACTIVEVOICEPROJECTLib_TLB;

type
  TfrmDirectSSControl = class(TForm)
    lstProgress: TListBox;
    Label1: TLabel;
    gbAttrs: TGroupBox;
    tbPitch: TTrackBar;
    tbSpeed: TTrackBar;
    tbVolume: TTrackBar;
    Label3: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    lblMinPitch: TLabel;
    lblMaxPitch: TLabel;
    lblMinSpeed: TLabel;
    lblMaxSpeed: TLabel;
    lblPitch: TLabel;
    lblSpeed: TLabel;
    lblMaxVolume: TLabel;
    lblMinVolume: TLabel;
    ImageList1: TImageList;
    btnSetAttrs: TButton;
    btnAbout: TButton;
    btnGeneral: TButton;
    btnLexicon: TButton;
    btnTranslate: TButton;
    cbEngines: TComboBox;
    Label7: TLabel;
    Label8: TLabel;
    DirectSS: TDirectSS;
    reText: TRichEdit;
    ToolBar1: TToolBar;
    btnPlay: TToolButton;
    btnPause: TToolButton;
    btnStop: TToolButton;
    memEnginePhonemes: TMemo;
    lstInfo: TListBox;
    Label2: TLabel;
    Label6: TLabel;
    procedure FormCreate(Sender: TObject);
    procedure cbEnginesChange(Sender: TObject);
    procedure btnAboutClick(Sender: TObject);
    procedure btnGeneralClick(Sender: TObject);
    procedure btnLexiconClick(Sender: TObject);
    procedure btnTranslateClick(Sender: TObject);
    procedure btnPlayClick(Sender: TObject);
    procedure btnPauseClick(Sender: TObject);
    procedure btnStopClick(Sender: TObject);
    procedure tbPitchChange(Sender: TObject);
    procedure tbSpeedChange(Sender: TObject);
    procedure btnSetAttrsClick(Sender: TObject);
    procedure DirectSSAttribChanged(Sender: TObject;
      which_attribute: Integer);
    procedure DirectSSAudioStart(Sender: TObject; hi, lo: Integer);
    procedure DirectSSAudioStop(Sender: TObject; hi, lo: Integer);
    procedure DirectSSVisual(Sender: TObject; timehi, timelo: Integer;
      Phoneme, EnginePhoneme: Smallint; hints: Integer; MouthHeight,
      bMouthWidth, bMouthUpturn, bJawOpen, TeethUpperVisible,
      TeethLowerVisible, TonguePosn, LipTension: Smallint);
    procedure DirectSSBookMark(Sender: TObject; hi, lo, MarkNum: Integer);
    procedure DirectSSTextDataDone(Sender: TObject; hi, lo,
      Flags: Integer);
    procedure DirectSSTextDataStarted(Sender: TObject; hi, lo: Integer);
    procedure DirectSSWordPosition(Sender: TObject; hi, lo,
      byteoffset: Integer);
    procedure DirectSSwarning(Sender: TObject; warning: Cardinal;
      const Details, Message: WideString);
    procedure DirectSSError(Sender: TObject; warning: Cardinal;
      const Details, Message: WideString);
    procedure DirectSSDebugging(Sender: TObject);
  private
    FOldPos: DWord;
    BeenPaused: Boolean;
    procedure CheckPitch;
    procedure CheckSpeed;
    procedure CheckVolume;
    procedure SetWordStyle(FirstChar: Integer; Styles: TFontStyles);
  public
    procedure Log(const Msg: String); overload;
    procedure Log(const Msg: String; const Args: array of const); overload;
  end;

var
  frmDirectSSControl: TfrmDirectSSControl;

implementation

{$R *.dfm}

{ TfrmDTTSTest }

procedure TfrmDirectSSControl.Log(const Msg: String);
begin
  if not Assigned(lstProgress) then
    Exit;
  lstProgress.Items.Add(Msg);
  lstProgress.ItemIndex := lstProgress.Items.Count - 1
end;

procedure TfrmDirectSSControl.Log(const Msg: String; const Args: array of const);
begin
  Log(Format(Msg, Args))
end;

procedure TfrmDirectSSControl.FormCreate(Sender: TObject);
var
  I: Integer;
begin
  SendMessage(lstProgress.Handle, LB_SETHORIZONTALEXTENT, Width, 0);
  DirectSS.Anchors := [akRight, akBottom];
  for I := 1 to DirectSS.CountEngines do
    cbEngines.Items.AddObject(DirectSS.ModeName(I), TObject(I));
  cbEngines.ItemIndex := cbEngines.Items.IndexOf(
    DirectSS.ModeName(DirectSS.CurrentMode));
  cbEngines.OnChange(cbEngines);
end;

procedure TfrmDirectSSControl.cbEnginesChange(Sender: TObject);
const
  Genders: array[GENDER_NEUTRAL..GENDER_MALE] of String =
    ('neutral', 'female', 'male');
var
  Buf: array[0..256] of Char;
begin
  DirectSS.CurrentMode :=
    Integer(cbEngines.Items.Objects[cbEngines.ItemIndex]);
  with lstInfo.Items do
  begin
    Clear;
    Add(Format('Manufacturer: %s',
      [DirectSS.MfgName(DirectSS.CurrentMode)]));
    Add(Format('Product Name: %s',
      [DirectSS.ProductName(DirectSS.CurrentMode)]));
    Add(Format('Engine ID: %s',
      [DirectSS.EngineID(DirectSS.CurrentMode)]));
    Add(Format('Mode ID: %s',
      [DirectSS.ModeID(DirectSS.CurrentMode)]));
    Add(Format('Mode Name: %s',
      [DirectSS.ModeName(DirectSS.CurrentMode)]));
    Win32Check(Bool(GetLocaleInfo(
      DirectSS.LanguageID(DirectSS.CurrentMode),
      LOCALE_SENGLANGUAGE, Buf, SizeOf(Buf))));
    Add(Format('Language ID: %s', [Buf]));
    Add(Format('Dialect: %s',
      [DirectSS.Dialect(DirectSS.CurrentMode)]));
    Add(Format('Gender: %s',
      [Genders[DirectSS.Gender(DirectSS.CurrentMode)]]));
    Add(Format('Voice Style: %s',
      [DirectSS.Style(DirectSS.CurrentMode)]));
    Add(Format('Speaker: %s',
      [DirectSS.Speaker(DirectSS.CurrentMode)]));
  end;
  CheckPitch;
  CheckSpeed;
  CheckVolume;
end;

procedure TfrmDirectSSControl.btnAboutClick(Sender: TObject);
begin
  DirectSS.AboutDlg(Handle, '');
end;

procedure TfrmDirectSSControl.btnGeneralClick(Sender: TObject);
begin
  DirectSS.GeneralDlg(Handle, '');
end;

procedure TfrmDirectSSControl.btnLexiconClick(Sender: TObject);
begin
  DirectSS.LexiconDlg(Handle, '');
end;

procedure TfrmDirectSSControl.btnTranslateClick(Sender: TObject);
begin
  DirectSS.TranslateDlg(Handle, '');
end;

procedure TfrmDirectSSControl.CheckPitch;
begin
  tbPitch.Min := DirectSS.MinPitch;
  lblMinPitch.Caption := IntToStr(tbPitch.Min);
  tbPitch.Max := DirectSS.MaxPitch;
  lblMaxPitch.Caption := IntToStr(tbPitch.Max);
  tbPitch.Position := DirectSS.Pitch;
  lblPitch.Caption := IntToStr(DirectSS.Pitch);
  tbPitch.Frequency := Trunc((tbPitch.Max - tbPitch.Min) / 10);
end;

procedure TfrmDirectSSControl.CheckSpeed;
begin
  tbSpeed.Min := DirectSS.MinSpeed;
  lblMinSpeed.Caption := IntToStr(tbSpeed.Min);
  tbSpeed.Max := DirectSS.MaxSpeed;
  lblMaxSpeed.Caption := IntToStr(tbSpeed.Max);
  tbSpeed.Position := DirectSS.Speed;
  lblSpeed.Caption := IntToStr(DirectSS.Speed);
  tbSpeed.Frequency := Trunc((tbSpeed.Max - tbSpeed.Min) / 10);
end;

procedure TfrmDirectSSControl.CheckVolume;
begin
  tbVolume.Min := DirectSS.MinVolumeRight;
  tbVolume.Max := DirectSS.MaxVolumeRight;
  tbVolume.Position := DirectSS.VolumeRight;
  tbVolume.Frequency := Trunc((tbVolume.Max - tbVolume.Min) / 10);
end;

procedure TfrmDirectSSControl.tbPitchChange(Sender: TObject);
begin
  lblPitch.Caption := IntToStr(tbPitch.Position)
end;

procedure TfrmDirectSSControl.tbSpeedChange(Sender: TObject);
begin
  lblSpeed.Caption := IntToStr(tbSpeed.Position)
end;

procedure TfrmDirectSSControl.btnSetAttrsClick(Sender: TObject);
begin
  DirectSS.Pitch := tbPitch.Position;
  lblPitch.Caption := IntToStr(tbPitch.Position);
  DirectSS.Speed := tbSpeed.Position;
  lblSpeed.Caption := IntToStr(tbSpeed.Position);
  DirectSS.VolumeRight := tbVolume.Position;
  DirectSS.VolumeLeft := tbVolume.Position;
end;

procedure TfrmDirectSSControl.btnPlayClick(Sender: TObject);
begin
    if not BeenPaused then
      DirectSS.Speak(reText.Text)
    else
    begin
      DirectSS.AudioResume;
      BeenPaused := False
    end
end;

procedure TfrmDirectSSControl.btnPauseClick(Sender: TObject);
begin
  if Bool(DirectSS.Speaking) then
  begin
    DirectSS.AudioPause;
    BeenPaused := True
  end
end;

procedure TfrmDirectSSControl.btnStopClick(Sender: TObject);
begin
  DirectSS.AudioReset;
end;

procedure TfrmDirectSSControl.SetWordStyle(FirstChar: Integer; Styles: TFontStyles);
var
  WordLen: Integer;
begin
  with reText do
  begin
    WordLen := 1;
    //Locate end of word
    while FirstChar + WordLen < Length(Text) do
    begin
      if not (Text[Succ(FirstChar) + WordLen] in ['a'..'z', 'A'..'Z', '0'..'9']) then
        Break;
      Inc(WordLen)
    end;
    Lines.BeginUpdate;
    try
      SelStart := FirstChar; //highlight word
      SelLength := WordLen;
      SelAttributes.Style := Styles; //apply requested style
      SelLength := 0;
    finally
     Lines.EndUpdate
    end
  end
end;

procedure TfrmDirectSSControl.DirectSSAttribChanged(Sender: TObject;
  which_attribute: Integer);
var
  S: String;
begin
  case which_attribute of
    TTSNSAC_REALTIME : S := 'Realtime';
    TTSNSAC_PITCH    : S := 'Pitch';
    TTSNSAC_SPEED    : S := 'Speed';
    TTSNSAC_VOLUME   : S := 'Volume';
  end;
  Log('Event AttribChanged: %s changed', [S]);
end;

function MakeInt64(Lo, Hi: Integer): Int64;
begin
  with LARGE_INTEGER(Result) do
  begin
    HighPart := Hi;
    LowPart := Lo
  end
end;

procedure TfrmDirectSSControl.DirectSSAudioStart(Sender: TObject; hi,
  lo: Integer);
begin
  Log('Event AudioStart: timestamp %d', [MakeInt64(lo, hi)]);
end;

procedure TfrmDirectSSControl.DirectSSAudioStop(Sender: TObject; hi, lo: Integer);
begin
  Log('Event AudioStop: timestamp %d', [MakeInt64(lo, hi)]);
end;

procedure TfrmDirectSSControl.DirectSSVisual(Sender: TObject; timehi,
  timelo: Integer; Phoneme, EnginePhoneme: Smallint; hints: Integer;
  MouthHeight, bMouthWidth, bMouthUpturn, bJawOpen, TeethUpperVisible,
  TeethLowerVisible, TonguePosn, LipTension: Smallint);
var
  Hint: String;
begin
  Hint := '';
  if Hints <> 0 then
  begin
    if Hints and TTSNSHINT_QUESTION <> 0 then
      Hint := 'Question ';
    if Hints and TTSNSHINT_STATEMENT <> 0 then
      Hint := Hint + 'Statement ';
    if Hints and TTSNSHINT_COMMAND <> 0 then
      Hint := Hint + 'Command ';
    if Hints and TTSNSHINT_EXCLAMATION <> 0 then
      Hint := Hint + 'Exclamation ';
    if Hints and TTSNSHINT_EMPHASIS <> 0 then
      Hint := Hint + 'Emphasis';
  end
  else
    Hint := 'none';
  Log('Event Visual: timestamp %d, hint = %s', [MakeInt64(timelo, timehi), Hint]);
  if Char(EnginePhoneme) <> #32 then
    memEnginePhonemes.Text := memEnginePhonemes.Text + Char(EnginePhoneme)
end;

procedure TfrmDirectSSControl.DirectSSBookMark(Sender: TObject; hi, lo,
  MarkNum: Integer);
begin
  Log('Event Bookmark: timestamp %d, bookmark %d',
    [MakeInt64(lo, hi), MarkNum]);
end;

procedure TfrmDirectSSControl.DirectSSTextDataDone(Sender: TObject; hi, lo,
  Flags: Integer);
const
  EndTypes: array[Boolean] of String = ('all text sent to audio device', 'audio was reset');
begin
  Log('Event TextDataDone: timestamp %d, %s',
    [MakeInt64(lo, hi), EndTypes[Flags = TTSBNS_ABORTED]]);
  SetWordStyle(FOldPos, []);
end;

procedure TfrmDirectSSControl.DirectSSTextDataStarted(Sender: TObject; hi,
  lo: Integer);
begin
  Log('Event TextDataStarted: timestamp %d', [MakeInt64(lo, hi)]);
  memEnginePhonemes.Clear
end;

procedure TfrmDirectSSControl.DirectSSWordPosition(Sender: TObject; hi, lo,
  byteoffset: Integer);
begin
  Log('Event WordPosition: timestamp %d, byte offset %d',
    [MakeInt64(lo, hi), ByteOffset]);
  SetWordStyle(FOldPos, []);
  //Byte offset is offset into Unicode buffer
  FOldPos := ByteOffset div SizeOf(WideChar);
  SetWordStyle(FOldPos, [fsBold]);
  if Trim(memEnginePhonemes.Text) <> '' then
    memEnginePhonemes.Text := memEnginePhonemes.Text + #32;
end;

procedure TfrmDirectSSControl.DirectSSwarning(Sender: TObject;
  warning: Cardinal; const Details, Message: WideString);
begin
  Log('Event Warning: %d, msg: %s, details: %s', [warning, Message, Details])
end;

procedure TfrmDirectSSControl.DirectSSError(Sender: TObject;
  warning: Cardinal; const Details, Message: WideString);
begin
  Log('Event Error: %d, msg: %s, details: %s', [warning, Message, Details])
end;

procedure TfrmDirectSSControl.DirectSSDebugging(Sender: TObject);
begin
  Log('Event Debugging')
end;

end.
