unit DirectTextToSpeechAPIU;
//Uses the JEDI Speech import unit and the low-Level DirectTextToSpeech API

{$Q-}

interface

uses
  Speech,
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, ImgList, ActnList, StdCtrls, ComCtrls, ToolWin;

type
  TTTSBufNotifySink = class;

  TfrmDirectTTSAPI = class(TForm)
    lstProgress: TListBox;
    Label1: TLabel;
    cbEngines: TComboBox;
    Label2: TLabel;
    lstEngineInfo: TListBox;
    gbAttrs: TGroupBox;
    tbPitch: TTrackBar;
    tbSpeed: TTrackBar;
    tbVolume: TTrackBar;
    Label3: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    lblMinPitch: TLabel;
    lblMaxPitch: TLabel;
    lblMinSpeed: TLabel;
    lblMaxSpeed: TLabel;
    lblPitch: TLabel;
    lblSpeed: TLabel;
    lblMaxVolume: TLabel;
    lblMinVolume: TLabel;
    ImageList1: TImageList;
    reText: TRichEdit;
    ToolBar1: TToolBar;
    btnPlay: TToolButton;
    btnPause: TToolButton;
    btnStop: TToolButton;
    btnSetAttrs: TButton;
    btnAbout: TButton;
    btnGeneral: TButton;
    btnLexicon: TButton;
    btnTranslate: TButton;
    Label6: TLabel;
    Label7: TLabel;
    memEnginePhonemes: TMemo;
    Label8: TLabel;
    chkStartStopEngine: TCheckBox;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure cbEnginesChange(Sender: TObject);
    procedure chkStartStopEngineClick(Sender: TObject);
    procedure btnAboutClick(Sender: TObject);
    procedure btnGeneralClick(Sender: TObject);
    procedure btnLexiconClick(Sender: TObject);
    procedure btnTranslateClick(Sender: TObject);
    procedure btnPlayClick(Sender: TObject);
    procedure btnPauseClick(Sender: TObject);
    procedure btnStopClick(Sender: TObject);
    procedure btnSetAttrsClick(Sender: TObject);
  private
    AMMD: IAudioMultiMediaDevice;
    TTSEnum: ITTSEnum;
    PModeInfo: PTTSModeInfo;
    TTSCentral: ITTSCentral;
    TTSDialogs: ITTSDialogs;
    TTSAttrs: ITTSAttributes;
    CurrentPitch: Word;
    CurrentSpeed, CurrentVolume: DWord;
    BeenPaused: Boolean;
    BufferSink: ITTSBufNotifySink;
    EngineSinkCookie: DWord;
    procedure CheckPitch;
    procedure CheckSpeed;
    procedure CheckVolume;
    procedure SetWordStyle(FirstChar: Integer; Styles: TFontStyles);
  public
    procedure Log(const Msg: String); overload;
    procedure Log(const Msg: String; const Args: array of const); overload;
  end;

  TTTSNotifySink = class(TInterfacedObject, ITTSNotifySink)
  private
    FForm: TfrmDirectTTSAPI;
  protected
    function AttribChanged(dwAttribute: DWORD): HResult; stdcall;
    function AudioStart(qTimeStamp: QWORD): HResult; stdcall;
    function AudioStop(qTimeStamp: QWORD): HResult; stdcall;
    function Visual(qTimeStamp: QWORD; cIPAPhoneme: Char;
      cEnginePhoneme: Char; dwHints: DWORD;
      apTTSMouth: PTTSMouth): HResult; stdcall;
  public
    constructor Create(Form: TfrmDirectTTSAPI);
  end;

  TTTSBufNotifySink = class(TInterfacedObject, ITTSBufNotifySink)
  private
    FForm: TfrmDirectTTSAPI;
    FOldPos: DWord;
    FinishedTalking: Boolean;
  protected
    function TextDataDone(qTimeStamp: QWORD; dwFlags: DWORD): HResult; stdcall;
    function TextDataStarted(qTimeStamp: QWORD): HResult; stdcall;
    function BookMark(qTimeStamp: QWORD; dwMarkNum: DWORD): HResult; stdcall;
    function WordPosition(qTimeStamp: QWORD; dwByteOffset: DWORD): HResult; stdcall;
  public
    constructor Create(Form: TfrmDirectTTSAPI);
  end;

var
  frmDirectTTSAPI: TfrmDirectTTSAPI;

implementation

{$R *.dfm}

uses
  ComObj, MMSystem;

{ TfrmDTTSTest }

procedure TfrmDirectTTSAPI.Log(const Msg: String);
begin
  if not Assigned(lstProgress) then
    Exit;
  lstProgress.Items.Add(Msg);
  lstProgress.ItemIndex := lstProgress.Items.Count - 1
end;

procedure TfrmDirectTTSAPI.Log(const Msg: String; const Args: array of const);
begin
  Log(Format(Msg, Args))
end;

procedure TfrmDirectTTSAPI.FormCreate(Sender: TObject);
var
  ModeInfo: TTTSModeInfo;
  PModeInfo: PTTSModeInfo;
  NumFound : DWord;
begin
  SendMessage(lstProgress.Handle, LB_SETHORIZONTALEXTENT, Width, 0);
  Log('About to connect to multimedia device');
  AMMD := CreateComObject(CLSID_MMAudioDest) as IAudioMultiMediaDevice;
  OleCheck(AMMD.DeviceNumSet(WAVE_MAPPER));
  Log('Connected to multimedia device');
  Log('About to enumerate speech engines');
  TTSEnum := CreateComObject(CLSID_TTSEnumerator) as ITTSEnum;
  OleCheck(TTSEnum.Reset);
  OleCheck(TTSEnum.Next(1, ModeInfo, @NumFound));
  While NumFound > 0 do
  begin
    New(PModeInfo);
    PModeInfo^ := ModeInfo;
    cbEngines.Items.AddObject(
      String(ModeInfo.szModeName), TObject(PModeInfo));
    OleCheck(TTSEnum.Next(1, ModeInfo, @NumFound));
  end;
  if cbEngines.Items.Count > 0 then
  begin
    cbEngines.ItemIndex := 0; //Select 1st engine
    cbEngines.OnChange(cbEngines); //& ensure OnChange triggers
  end;
  Log('Enumerated speech engines');
end;

procedure TfrmDirectTTSAPI.FormDestroy(Sender: TObject);
var
  I: Integer;
begin
  chkStartStopEngine.Checked := False;
  //Clean up TTTSModeInfo memory for each speech engine
  for I := 0 to cbEngines.Items.Count - 1 do
    Dispose(PTTSModeInfo(cbEngines.Items.Objects[I]));
end;

procedure TfrmDirectTTSAPI.cbEnginesChange(Sender: TObject);
const
  Genders: array[GENDER_NEUTRAL..GENDER_MALE] of String =
    ('Neutral', 'Female', 'Male');
begin
  with lstEngineInfo.Items do
  begin
    Clear;
    PModeInfo := PTTSModeInfo(cbEngines.Items.Objects[cbEngines.ItemIndex]);
    Add(Format('Name:'#9'%s', [PModeInfo.szModeName]));
    Add(Format('Language:'#9'%s', [PModeInfo.Language.szDialect]));
    Add(Format('Manufacturer:'#9'%s', [PModeInfo.szMfgName]));
    Add(Format('Product:'#9'%s', [PModeInfo.szProductName]));
    Add(Format('Mode ID:'#9'%s', [GUIDToString(PModeInfo.gModeID)]));
    Add(Format('Engine ID:'#9'%s', [GUIDToString(PModeInfo.gEngineID)]));
    Add(Format('Speaker:'#9'%s', [PModeInfo.szSpeaker]));
    Add(Format('Style:'#9'%s', [PModeInfo.szStyle]));
    Add(Format('Gender:'#9'%s', [Genders[PModeInfo.wGender]]));
    case PModeInfo.wAge of
      TTSAGE_BABY      : Add('Age:'#9'Baby');
      TTSAGE_TODDLER   : Add('Age:'#9'Toddler');
      TTSAGE_CHILD     : Add('Age:'#9'Child');
      TTSAGE_ADOLESCENT: Add('Age:'#9'Adolescent');
      TTSAGE_ADULT     : Add('Age:'#9'Adult');
      TTSAGE_ELDERLY   : Add('Age:'#9'Elderly');
    end;
  end
end;

procedure TfrmDirectTTSAPI.chkStartStopEngineClick(Sender: TObject);
var
  TTSNotifySink: ITTSNotifySink;
begin
  if chkStartStopEngine.Checked then
  begin
    Log('About to start selected speech engine');
    OleCheck(TTSEnum.Select(PModeInfo.gModeID, TTSCentral, AMMD));
    btnPlay.Enabled := True;
    btnPause.Enabled := True;
    btnStop.Enabled := True;
    Log('Started selected speech engine');
    Log('About to make dialogs available');
    if PModeInfo.dwInterfaces and TTSI_ITTSDialogs > 0 then
    begin
      TTSDialogs := TTSCentral as ITTSDialogs;
      btnAbout.Enabled := True;
      btnGeneral.Enabled := True;
      btnLexicon.Enabled := True;
      btnTranslate.Enabled := True;
      Log('Dialogs are available');
    end
    else
      Log('Dialogs are not supported');
    Log('About to check speech attributes');
    if PModeInfo.dwInterfaces and TTSI_ITTSATTRIBUTES > 0 then
    begin
      TTSAttrs := TTSCentral as ITTSAttributes;
      CheckPitch;
      CheckSpeed;
      CheckVolume;
      Log('Attributes checked');
      btnSetAttrs.Enabled := True;
      gbAttrs.Enabled := True
    end
    else
      Log('Attributes interface not supported');
    Log('About to create and register engine event sink');
    TTSNotifySink := TTTSNotifySink.Create(Self);
    OleCheck(TTSCentral.Register(Pointer(TTSNotifySink),
      ITTSNotifySink, EngineSinkCookie));
    Log('Registered engine event sink');
    Log('About to create buffer event sink');
    BufferSink := TTTSBufNotifySink.Create(Self);
    Log('Created engine buffer sink')
  end
  else
  begin
    Log('About to unregister engine event sink');
    OleCheck(TTSCentral.UnRegister(EngineSinkCookie));
    Log('Unregistered engine event sink');
    Log('About to destroy buffer event sink');
    BufferSink := nil;
    Log('Destroyed engine buffer sink');
    Log('About to stop speech engine');
    TTSCentral := nil;
    btnPlay.Enabled := False;
    btnPause.Enabled := False;
    btnStop.Enabled := False;
    TTSDialogs := nil;
    btnAbout.Enabled := False;
    btnGeneral.Enabled := False;
    btnLexicon.Enabled := False;
    btnTranslate.Enabled := False;
    TTSAttrs := nil;
    btnSetAttrs.Enabled := False;
    gbAttrs.Enabled := False;
    Log('Speech engine stopped');
  end
end;

procedure TfrmDirectTTSAPI.btnAboutClick(Sender: TObject);
begin
  OleCheck(TTSDialogs.AboutDlg(Handle, nil));
end;

procedure TfrmDirectTTSAPI.btnGeneralClick(Sender: TObject);
begin
  OleCheck(TTSDialogs.GeneralDlg(Handle, nil));
end;

procedure TfrmDirectTTSAPI.btnLexiconClick(Sender: TObject);
begin
  OleCheck(TTSDialogs.LexiconDlg(Handle, nil));
end;

procedure TfrmDirectTTSAPI.btnTranslateClick(Sender: TObject);
begin
  OleCheck(TTSDialogs.TranslateDlg(Handle, nil));
end;

procedure TfrmDirectTTSAPI.CheckPitch;
var
  MinPitch, MaxPitch: Word;
begin
  tbPitch.Enabled := False;
  if PModeInfo.dwFeatures and TTSFEATURE_PITCH <> 0 then
  begin
    Log('Engine claims to support voice pitch');
    if TTSAttrs.PitchGet(CurrentPitch) <> TTSERR_NOTSUPPORTED then
    begin
      Log('Retrieved current pitch');
      if TTSAttrs.PitchSet(1) <> TTSERR_NOTSUPPORTED then
      begin
        Log('Pitch can be set');
        OleCheck(TTSAttrs.PitchSet(TTSATTR_MINPITCH));
        OleCheck(TTSAttrs.PitchGet(MinPitch));
        tbPitch.Min := MinPitch;
        lblMinPitch.Caption := IntToStr(MinPitch);
        OleCheck(TTSAttrs.PitchSet(TTSATTR_MAXPITCH));
        OleCheck(TTSAttrs.PitchGet(MaxPitch));
        tbPitch.Max := MaxPitch;
        lblMaxPitch.Caption := IntToStr(MaxPitch);
        OleCheck(TTSAttrs.PitchSet(CurrentPitch));
        tbPitch.Position := CurrentPitch;
        lblPitch.Caption := IntToStr(CurrentPitch);
        tbPitch.Frequency := Trunc((MaxPitch - MinPitch) / 10);
        tbPitch.Enabled := True;
      end
      else
        Log('Pitch cannot be set')
    end
    else
      Log('Pitch retrieval failed')
  end
  else
    Log('Engine does not support voice pitch');
end;

procedure TfrmDirectTTSAPI.CheckSpeed;
var
  MinSpeed, MaxSpeed: DWord;
begin
  tbSpeed.Enabled := False;
  if PModeInfo.dwFeatures and TTSFEATURE_SPEED <> 0 then
  begin
    Log('Engine claims to support voice speed');
    if TTSAttrs.SpeedGet(CurrentSpeed) <> TTSERR_NOTSUPPORTED then
    begin
      Log('Retrieved current speed');
      if TTSAttrs.SpeedSet(1) <> TTSERR_NOTSUPPORTED then
      begin
        Log('Speed can be set');
        OleCheck(TTSAttrs.SpeedSet(TTSATTR_MINSPEED));
        OleCheck(TTSAttrs.SpeedGet(MinSpeed));
        tbSpeed.Min := MinSpeed;
        lblMinSpeed.Caption := IntToStr(MinSpeed);
        OleCheck(TTSAttrs.SpeedSet(TTSATTR_MAXSPEED));
        OleCheck(TTSAttrs.SpeedGet(MaxSpeed));
        tbSpeed.Max := MaxSpeed;
        lblMaxSpeed.Caption := IntToStr(MaxSpeed);
        OleCheck(TTSAttrs.SpeedSet(CurrentSpeed));
        tbSpeed.Position := CurrentSpeed;
        lblSpeed.Caption := IntToStr(CurrentSpeed);
        tbSpeed.Frequency := Trunc((MaxSpeed - MinSpeed) / 10);
        tbSpeed.Enabled := True;
      end
      else
        Log('Speed cannot be set')
    end
    else
      Log('Speed retrieval failed')
  end
  else
    Log('Engine does not support voice speed');
end;

procedure TfrmDirectTTSAPI.CheckVolume;
var
  MinVolume, MaxVolume: DWord;
begin
  tbVolume.Enabled := False;
  if PModeInfo.dwFeatures and TTSFEATURE_VOLUME <> 0 then
  begin
    Log('Engine claims to support volume');
    if TTSAttrs.VolumeGet(CurrentVolume) <> TTSERR_NOTSUPPORTED then
    begin
      Log('Retrieved current volume');
      if TTSAttrs.VolumeSet(1) <> TTSERR_NOTSUPPORTED then
      begin
        Log('Volume can be set');
        OleCheck(TTSAttrs.VolumeSet(TTSATTR_MINVOLUME));
        OleCheck(TTSAttrs.VolumeGet(MinVolume));
        tbVolume.Min := MinVolume div 2;
        OleCheck(TTSAttrs.VolumeSet(TTSATTR_MAXVOLUME));
        OleCheck(TTSAttrs.VolumeGet(MaxVolume));
        tbVolume.Max := MaxVolume div 2;
        OleCheck(TTSAttrs.VolumeSet(CurrentVolume));
        tbVolume.Position := CurrentVolume div 2;
        tbVolume.Frequency := Trunc((MaxVolume - MinVolume) / 20);
        tbVolume.Enabled := True;
      end
      else
        Log('Volume cannot be set')
    end
    else
      Log('Volume retrieval failed')
  end
  else
    Log('Engine does not support volume');
end;

procedure TfrmDirectTTSAPI.btnPlayClick(Sender: TObject);
var
  SData: TSData;
begin
  if not BeenPaused then
  begin
    SData.dwSize := Succ(Length(reText.Text));
    SData.pData := PChar(reText.Text);
    OleCheck(TTSCentral.TextData(CHARSET_TEXT, 0, SData,
      Pointer(BufferSink), ITTSBufNotifySink));
  end
  else
  begin
    OleCheck(TTSCentral.AudioResume);
    BeenPaused := False
  end
end;

procedure TfrmDirectTTSAPI.btnPauseClick(Sender: TObject);
begin
  OleCheck(TTSCentral.AudioPause);
  BeenPaused := True
end;

procedure TfrmDirectTTSAPI.btnStopClick(Sender: TObject);
begin
  OleCheck(TTSCentral.AudioReset);
end;

procedure TfrmDirectTTSAPI.btnSetAttrsClick(Sender: TObject);
begin
  if tbPitch.Enabled then
  begin
    TTSAttrs.PitchSet(tbPitch.Position);
    lblPitch.Caption := IntToStr(tbPitch.Position)
  end;
  if tbSpeed.Enabled then
  begin
    TTSAttrs.SpeedSet(tbSpeed.Position);
    lblSpeed.Caption := IntToStr(tbSpeed.Position)
  end;
  if tbVolume.Enabled then
    TTSAttrs.VolumeSet(DWord(tbVolume.Position) * 2);
end;

procedure TfrmDirectTTSAPI.SetWordStyle(FirstChar: Integer; Styles: TFontStyles);
var
  WordLen: Integer;
begin
  with reText do
  begin
    WordLen := 1;
    while FirstChar + WordLen < Length(Text) do //Locate end of word
    begin
      if not (Text[Succ(FirstChar) + WordLen] in
              ['a'..'z', 'A'..'Z', '0'..'9']) then
        Break;
      Inc(WordLen)
    end;
    Lines.BeginUpdate;
    try
      SelStart := FirstChar; //highlight word
      SelLength := WordLen;
      SelAttributes.Style := Styles; //apply requested style
      SelLength := 0; //unhighlight word
    finally
      Lines.EndUpdate
    end
  end
end;

{ TTTSNotifySink }

constructor TTTSNotifySink.Create(Form: TfrmDirectTTSAPI);
begin
  inherited Create;
  FForm := Form
end;

function TTTSNotifySink.AttribChanged(dwAttribute: DWORD): HResult;
var
  S: String;
begin
  Result := S_OK;
  case dwAttribute of
    TTSNSAC_REALTIME : S := 'Realtime';
    TTSNSAC_PITCH    : S := 'Pitch';
    TTSNSAC_SPEED    : S := 'Speed';
    TTSNSAC_VOLUME   : S := 'Volume';
  end;
  FForm.Log('Engine Event AttribChanged: %s changed', [S]);
end;

function TTTSNotifySink.AudioStart(qTimeStamp: QWORD): HResult;
begin
  Result := S_OK;
  FForm.Log('Engine Event AudioStart: timestamp %d', [qTimeStamp]);
end;

function TTTSNotifySink.AudioStop(qTimeStamp: QWORD): HResult;
begin
  Result := S_OK;
  FForm.Log('Engine Event AudioStop: timestamp %d', [qTimeStamp]);
end;

function TTTSNotifySink.Visual(qTimeStamp: QWORD; cIPAPhoneme,
  cEnginePhoneme: Char; dwHints: DWORD; apTTSMouth: PTTSMouth): HResult;
var
  Hint: String;
begin
  Result := S_OK;
  if cEnginePhoneme = '@' then
    Exit;
  Hint := '';
  if dwHints <> 0 then
  begin
    if dwHints and TTSNSHINT_QUESTION <> 0 then
      Hint := 'Question ';
    if dwHints and TTSNSHINT_STATEMENT <> 0 then
      Hint := Hint + 'Statement ';
    if dwHints and TTSNSHINT_COMMAND <> 0 then
      Hint := Hint + 'Command ';
    if dwHints and TTSNSHINT_EXCLAMATION <> 0 then
      Hint := Hint + 'Exclamation ';
    if dwHints and TTSNSHINT_EMPHASIS <> 0 then
      Hint := Hint + 'Emphasis';
  end
  else
    Hint := 'none';
  FForm.Log('Engine Event Visual: timestamp %d, hint = %s', [qTimeStamp, Hint]);
  //Assuming the buffer events are firing, we can write out the engine
  //phonemes and the WordPosition event will insert the spaces between words
  if cEnginePhoneme <> #32 then
    FForm.memEnginePhonemes.Text := FForm.memEnginePhonemes.Text + cEnginePhoneme
end;

{ TTTSBufNotifySink }

constructor TTTSBufNotifySink.Create(Form: TfrmDirectTTSAPI);
begin
  inherited Create;
  FForm := Form
end;

function TTTSBufNotifySink.BookMark(qTimeStamp: QWORD;
  dwMarkNum: DWORD): HResult;
begin
  Result := S_OK;
  FForm.Log('Buffer Event Bookmark: timestamp %d, bookmark %d',
    [qTimeStamp, dwMarkNum]);
end;

function TTTSBufNotifySink.TextDataDone(qTimeStamp: QWORD;
  dwFlags: DWORD): HResult;
const
  EndTypes: array[Boolean] of String =
    ('all text sent to audio device', 'audio was reset');
begin
  Result := S_OK;
  FForm.Log('Buffer Event TextDataDone: timestamp %d, %s',
    [qTimeStamp, EndTypes[dwFlags = TTSBNS_ABORTED]]);
  FForm.SetWordStyle(FOldPos, []);
  FinishedTalking := True;
end;

function TTTSBufNotifySink.TextDataStarted(qTimeStamp: QWORD): HResult;
begin
  Result := S_OK;
  FForm.Log('Buffer Event TextDataStarted: timestamp %d', [qTimeStamp]);
  if FinishedTalking then
    FForm.memEnginePhonemes.Clear;
  FinishedTalking := False;
end;

function TTTSBufNotifySink.WordPosition(qTimeStamp: QWORD;
  dwByteOffset: DWORD): HResult;
begin
  Result := S_OK;
  FForm.Log('Buffer Event WordPosition: timestamp %d, byte offset %d',
    [qTimeStamp, dwByteOffset]);
  FForm.SetWordStyle(FOldPos, []);
  //Byte offset is offset into Ansi buffer
  FOldPos := dwByteOffset;
  FForm.SetWordStyle(FOldPos, [fsBold]);
  //Assuming the engine events are firing, we can insert spaces
  //between the engine phoneme words made by the Visual event
  if Trim(FForm.memEnginePhonemes.Text) <> '' then
    FForm.memEnginePhonemes.Text := FForm.memEnginePhonemes.Text + #32;
end;

end.
