unit TextToSpeechControlU;

//Uses the high level Microsoft TextToSpeech ActiveX to
//simplify coding and to provide animation

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, ImgList, StdCtrls, ComCtrls, ToolWin, OleCtrls, HTTSLib_TLB;

type
  TfrmTextToSpeechControl = class(TForm)
    lstProgress: TListBox;
    Label1: TLabel;
    ImageList1: TImageList;
    memText: TMemo;
    ToolBar1: TToolBar;
    btnPlay: TToolButton;
    btnPause: TToolButton;
    btnStop: TToolButton;
    Label7: TLabel;
    TextToSpeech: TTextToSpeech;
    btnAbout: TButton;
    btnGeneral: TButton;
    btnLexicon: TButton;
    btnTranslate: TButton;
    Label8: TLabel;
    memEnginePhonemes: TMemo;
    procedure FormCreate(Sender: TObject);
    procedure btnPlayClick(Sender: TObject);
    procedure btnPauseClick(Sender: TObject);
    procedure btnStopClick(Sender: TObject);
    procedure btnAboutClick(Sender: TObject);
    procedure btnGeneralClick(Sender: TObject);
    procedure btnLexiconClick(Sender: TObject);
    procedure btnTranslateClick(Sender: TObject);
    procedure TextToSpeechAttribChanged(Sender: TObject; attrib: Integer);
    procedure TextToSpeechSpeak(Sender: TObject; const Text,
      App: WideString; thetype: Integer);
    procedure TextToSpeechSpeakingStarted(Sender: TObject);
    procedure TextToSpeechSpeakingDone(Sender: TObject);
    procedure TextToSpeechVisual(Sender: TObject; Phoneme,
      EnginePhoneme: Smallint; hints: Integer; MouthHeight, bMouthWidth,
      bMouthUpturn, bJawOpen, TeethUpperVisible, TeethLowerVisible,
      TonguePosn, LipTension: Smallint);
  private
    BeenPaused: Boolean;
  public
    procedure Log(const Msg: String); overload;
    procedure Log(const Msg: String; const Args: array of const); overload;
  end;

var
  frmTextToSpeechControl: TfrmTextToSpeechControl;

implementation

{$R *.dfm}

uses
  Speech;

{ TfrmVoiceTextControl }

procedure TfrmTextToSpeechControl.Log(const Msg: String);
begin
  if not Assigned(lstProgress) then
    Exit;
  lstProgress.Items.Add(Msg);
  lstProgress.ItemIndex := lstProgress.Items.Count - 1
end;

procedure TfrmTextToSpeechControl.Log(const Msg: String; const Args: array of const);
begin
  Log(Format(Msg, Args))
end;

procedure TfrmTextToSpeechControl.FormCreate(Sender: TObject);
begin
  TextToSpeech.Anchors := [akRight, akBottom]
end;

procedure TfrmTextToSpeechControl.btnPlayClick(Sender: TObject);
begin
  if not BeenPaused then
    TextToSpeech.Speak(memText.Text)
  else
  begin
    TextToSpeech.Resume;
    BeenPaused := False
  end
end;

procedure TfrmTextToSpeechControl.btnPauseClick(Sender: TObject);
begin
  if Bool(TextToSpeech.IsSpeaking) then
  begin
    TextToSpeech.Pause;
    BeenPaused := True
  end
end;

procedure TfrmTextToSpeechControl.btnStopClick(Sender: TObject);
begin
  TextToSpeech.StopSpeaking;
end;

procedure TfrmTextToSpeechControl.btnAboutClick(Sender: TObject);
begin
  TextToSpeech.AboutDlg(Handle, '');
end;

procedure TfrmTextToSpeechControl.btnGeneralClick(Sender: TObject);
begin
  TextToSpeech.GeneralDlg(Handle, '');
end;

procedure TfrmTextToSpeechControl.btnLexiconClick(Sender: TObject);
begin
  TextToSpeech.LexiconDlg(Handle, '');
end;

procedure TfrmTextToSpeechControl.btnTranslateClick(Sender: TObject);
begin
  TextToSpeech.TranslateDlg(Handle, '');
end;

procedure TfrmTextToSpeechControl.TextToSpeechAttribChanged(Sender: TObject;
  attrib: Integer);
var
  S: String;
begin
  case attrib of
    TTSNSAC_REALTIME : S := 'Realtime';
    TTSNSAC_PITCH    : S := 'Pitch';
    TTSNSAC_SPEED    : S := 'Speed';
    TTSNSAC_VOLUME   : S := 'Volume';
  end;
  Log('OnAttribChanged: %s changed', [S]);
end;

procedure TfrmTextToSpeechControl.TextToSpeechSpeak(Sender: TObject; const Text,
  App: WideString; thetype: Integer);
begin
  Log('OnSpeak');
  memEnginePhonemes.Clear
end;

procedure TfrmTextToSpeechControl.TextToSpeechSpeakingStarted(Sender: TObject);
begin
  Log('OnSpeakingStarted');
end;

procedure TfrmTextToSpeechControl.TextToSpeechSpeakingDone(Sender: TObject);
begin
  Log('OnSpeakingDone')
end;

procedure TfrmTextToSpeechControl.TextToSpeechVisual(Sender: TObject;
  Phoneme, EnginePhoneme: Smallint; hints: Integer; MouthHeight,
  bMouthWidth, bMouthUpturn, bJawOpen, TeethUpperVisible,
  TeethLowerVisible, TonguePosn, LipTension: Smallint);
var
  Hint: String;
begin
  Hint := '';
  if hints <> 0 then
  begin
    if hints and TTSNSHINT_QUESTION <> 0 then
      Hint := 'Question ';
    if hints and TTSNSHINT_STATEMENT <> 0 then
      Hint := Hint + 'Statement ';
    if hints and TTSNSHINT_COMMAND <> 0 then
      Hint := Hint + 'Command ';
    if hints and TTSNSHINT_EXCLAMATION <> 0 then
      Hint := Hint + 'Exclamation ';
    if hints and TTSNSHINT_EMPHASIS <> 0 then
      Hint := Hint + 'Emphasis';
  end
  else
    Hint := 'none';
  Log('OnVisual: hint = %s', [Hint]);
  if Char(EnginePhoneme) <> #32 then
    memEnginePhonemes.Text := memEnginePhonemes.Text + Char(EnginePhoneme)
end;

end.
